open StdLabels
module Report = Qsp_syntax.Report

type result = Report.t list [@@deriving show]
type filters = { level : Report.level option }

module Args = struct
  let input_files = ref []
  let usage = "qsp_parser input_file"
  let anon_fun filename = input_files := filename :: !input_files
  let level_value = ref None
  let reset_line = ref false

  let level : string -> unit =
   fun str_level ->
    match Report.level_of_string str_level with
    | Ok level_ -> level_value := Some level_
    | Error e ->
        print_endline e;
        exit 1

  let speclist =
    [
      ( "--version",
        Arg.Unit
          (fun () ->
            Printf.printf "Version %s\n" Tools.Git_hash.revision;
            exit 0),
        "Display the version of the application and exit" );
      ("--level", Arg.String level, "Message level [debug, warn, error]");
      ( "--global",
        Arg.Set reset_line,
        "Each line is refered from the begining of the file and not the \
         location" );
    ]

  let parse () =
    let () = Arg.parse speclist anon_fun usage in
    let filters = { level = !level_value } in
    (!input_files, !reset_line, filters)
end

(** Filter the results given by the analysis *)
let filter_report : filters -> Report.t list -> Report.t -> Report.t list =
 fun filters reports r ->
  let is_ok =
    match filters.level with
    | None -> true
    | Some level -> Report.level_to_enum level >= Report.level_to_enum r.level
  in

  match is_ok with true -> r :: reports | _ -> reports

(** Read the source file until getting a report (the whole location has been
    read properly), or until the first syntax error. 
    *)
let parse_location : Qparser.Lexbuf.t -> filters -> unit =
 fun lexbuf filters ->
  let result = Qparser.Analyzer.parse (module Qsp_syntax.Type_of) lexbuf in

  let result =
    Result.map
      (fun f -> List.fold_left (f []) ~init:[] ~f:(filter_report filters))
      result
  in
  match result with
  | Ok report -> (
      (* Display the result *)
      match report with
      | [] -> ()
      | _ ->
          let start_position, _ = Qparser.Lexbuf.positions lexbuf in
          Format.fprintf Format.std_formatter "Location@ %s@;@[%a@]@."
            start_position.Lexing.pos_fname pp_result report;
          ())
  | Error e ->
      let start_position, _ = Qparser.Lexbuf.positions lexbuf in
      Format.fprintf Format.std_formatter "Location@ %s@;@[%a]@."
        start_position.Lexing.pos_fname Report.pp e

let () =
  let file_names, reset_line, filters = Args.parse () in
  let file_name = List.hd file_names in

  let ic = Stdlib.open_in_bin file_name in

  (*let lexer = Lexing.from_channel ~with_positions:true ic in*)
  let lexer =
    match Filename.extension file_name with
    | ".qsrc" -> Sedlexing.Utf8.from_channel ic
    | ".txt" -> Sedlexing.Utf16.from_channel ic (Some Little_endian)
    | _ -> raise (Failure "unknown extension")
  in
  let lexer = Qparser.Lexbuf.from_lexbuf ~reset_line lexer in

  let () =
    try
      while true do
        parse_location lexer filters
      done
    with Qparser.Lexer.EOF -> ()
  in
  let () =
    match Sys.os_type with
    | "Win32" ->
        print_endline "Press <Enter> to terminate";
        ignore @@ read_line ()
    | _ -> ()
  in
  ()
